@@ -78,18 +78,18 @@ clientDumpList (ScreenInfo *screen_info)
 #endif
 
 static ClientPair
-clientGetTopMostFocusable (ScreenInfo *screen_info, int layer, GList * exclude_list)
+clientGetTopMostFocusable (ScreenInfo *screen_info, guint layer, GList * exclude_list)
 {
     ClientPair top_client;
     Client *c;
-    GList *index;
+    GList *list;
 
     TRACE ("entering clientGetTopMostFocusable");
 
     top_client.prefered = top_client.highest = NULL;
-    for (index = screen_info->windows_stack; index; index = g_list_next (index))
+    for (list = screen_info->windows_stack; list; list = g_list_next (list))
     {
-        c = (Client *) index->data;
+        c = (Client *) list->data;
         TRACE ("*** stack window \"%s\" (0x%lx), layer %i", c->name,
             c->window, (int) c->win_layer);
 
@@ -106,7 +106,7 @@ clientGetTopMostFocusable (ScreenInfo *screen_info, int layer, GList * exclude_l
             }
             else if (FLAG_TEST (c->xfwm_flags, XFWM_FLAG_VISIBLE))
             {
-                if (clientSelectMask (c, 0, WINDOW_REGULAR_FOCUSABLE))
+                if (clientSelectMask (c, NULL, 0, WINDOW_REGULAR_FOCUSABLE))
                 {
                     top_client.prefered = c;
                 }
@@ -119,7 +119,7 @@ clientGetTopMostFocusable (ScreenInfo *screen_info, int layer, GList * exclude_l
 }
 
 void
-clientFocusTop (ScreenInfo *screen_info, int layer, guint32 timestamp)
+clientFocusTop (ScreenInfo *screen_info, guint layer, guint32 timestamp)
 {
     ClientPair top_client;
     DisplayInfo *display_info;
@@ -162,19 +162,34 @@ clientFocusNew(Client * c)
     {
         give_focus = FALSE;
     }
+    else if (FLAG_TEST (c->flags, CLIENT_FLAG_HAS_STARTUP_TIME | CLIENT_FLAG_HAS_USER_TIME) && (c->user_time == (guint32) 0))
+    {
+        TRACE ("Given startup time is nil, not focusing \"%s\"", c->name);
+        give_focus = FALSE;
+        prevented = FALSE;
+    }
     else if ((client_focus) && (prevent_focus_stealing))
     {
-        if (FLAG_TEST (c->flags, CLIENT_FLAG_HAS_STARTUP_TIME) && (c->user_time == CurrentTime))
+        if (client_focus->win_layer > c->win_layer)
         {
-            TRACE ("Given startup time is 0, not focusing");
+            TRACE ("Not focusing \"%s\" because the current focused window is on a upper layer", c->name);
             give_focus = FALSE;
             prevented = TRUE;
         }
+        else if (client_focus->win_layer < c->win_layer)
+        {
+            /* We short don't use focus stealing prevention against upper layers */
+            TRACE ("Ignoring startup prevention because the current focused window is on a lower layer");
+            give_focus = TRUE;
+            prevented = FALSE;
+        }
         else if (FLAG_TEST (c->flags, CLIENT_FLAG_HAS_STARTUP_TIME | CLIENT_FLAG_HAS_USER_TIME))
         {
+            TRACE ("Current time is %u, time for \"%s\" is %u",
+                   (unsigned int) client_focus->user_time,
+                   c->name, (unsigned int) c->user_time);
             if (TIMESTAMP_IS_BEFORE (c->user_time, client_focus->user_time))
             {
-                TRACE ("Current %u, new %u", (unsigned int) client_focus->user_time, (unsigned int) c->user_time);
                 give_focus = FALSE;
                 prevented = TRUE;
             }
@@ -199,17 +214,16 @@ clientFocusNew(Client * c)
     {
         Client *c2 = clientGetFocus();
 
-        if ((c2 != NULL) && (c2->win_layer == c->win_layer))
+        /*
+         * Place windows under the currently focused only if focus
+         * stealing prevention had prevented the focus transition,
+         * otherwise, leave the unfocused window on top.
+         */
+        if ((c2 != NULL) && (c2->win_layer == c->win_layer) && prevented)
         {
-            if (prevented)
-            {
-                TRACE ("clientFocusNew: Setting WM_STATE_DEMANDS_ATTENTION flag on \"%s\" (0x%lx)", c->name, c->window);
-                FLAG_SET (c->flags, CLIENT_FLAG_DEMANDS_ATTENTION);
-            }
             clientSortRing(c);
             clientLower (c, c2->frame);
             clientSortRing(c2);
-            clientSetOpacity (c, c->opacity, 0, 0);
         }
         else
         {
@@ -217,6 +231,12 @@ clientFocusNew(Client * c)
             clientSortRing(c);
         }
 
+        if (prevented)
+        {
+            TRACE ("clientFocusNew: Setting WM_STATE_DEMANDS_ATTENTION flag on \"%s\" (0x%lx)", c->name, c->window);
+            FLAG_SET (c->flags, CLIENT_FLAG_DEMANDS_ATTENTION);
+        }
+
         clientShow (c, TRUE);
         clientGrabMouseButton (c);
         clientSetNetState (c);
@@ -226,30 +246,36 @@ clientFocusNew(Client * c)
 }
 
 gboolean
-clientSelectMask (Client * c, int mask, int type)
+clientSelectMask (Client * c, Client *other, guint mask, guint type)
 {
     g_return_val_if_fail (c != NULL, FALSE);
     TRACE ("entering clientSelectMask");
 
-    if ((!clientAcceptFocus (c)) && !(mask & INCLUDE_SKIP_FOCUS))
+    if ((mask & SEARCH_SAME_APPLICATION) && !clientSameApplication (c, other))
     {
         return FALSE;
     }
-    if (FLAG_TEST (c->flags, CLIENT_FLAG_ICONIFIED) && !(mask & INCLUDE_HIDDEN))
+    if ((mask & SEARCH_DIFFERENT_APPLICATION) && clientSameApplication (c, other))
     {
         return FALSE;
     }
-    if (FLAG_TEST (c->flags, CLIENT_FLAG_SKIP_PAGER)
-        && !(mask & INCLUDE_SKIP_PAGER))
+    if (!(mask & SEARCH_INCLUDE_SKIP_FOCUS) && !clientAcceptFocus (c))
     {
         return FALSE;
     }
-    if (FLAG_TEST (c->flags, CLIENT_FLAG_SKIP_TASKBAR)
-        && !(mask & INCLUDE_SKIP_TASKBAR))
+    if (!(mask & SEARCH_INCLUDE_HIDDEN) && FLAG_TEST (c->flags, CLIENT_FLAG_ICONIFIED))
     {
         return FALSE;
     }
-    if ((c->win_workspace != c->screen_info->current_ws) && !(mask & INCLUDE_ALL_WORKSPACES))
+    if (!(mask & SEARCH_INCLUDE_SKIP_PAGER) && FLAG_TEST (c->flags, CLIENT_FLAG_SKIP_PAGER))
+    {
+        return FALSE;
+    }
+    if (!(mask & SEARCH_INCLUDE_SKIP_TASKBAR) && FLAG_TEST (c->flags, CLIENT_FLAG_SKIP_TASKBAR))
+    {
+        return FALSE;
+    }
+    if (!(mask & SEARCH_INCLUDE_ALL_WORKSPACES) && (c->win_workspace != c->screen_info->current_ws))
     {
         return FALSE;
     }
@@ -262,7 +288,7 @@ clientSelectMask (Client * c, int mask, int type)
 }
 
 Client *
-clientGetNext (Client * c, int mask)
+clientGetNext (Client * c, guint mask, guint type)
 {
     Client *c2;
     unsigned int i;
@@ -275,7 +301,7 @@ clientGetNext (Client * c, int mask)
         for (c2 = c->next, i = 0; (c2) && (i < screen_info->client_count - 1);
             c2 = c2->next, i++)
         {
-            if (clientSelectMask (c2, mask, WINDOW_REGULAR_FOCUSABLE))
+            if (clientSelectMask (c2, c, mask, type))
             {
                 return c2;
             }
@@ -285,7 +311,7 @@ clientGetNext (Client * c, int mask)
 }
 
 Client *
-clientGetPrevious (Client * c, int mask)
+clientGetPrevious (Client * c, guint mask, guint type)
 {
     Client *c2;
     unsigned int i;
@@ -298,7 +324,7 @@ clientGetPrevious (Client * c, int mask)
         for (c2 = c->prev, i = 0; (c2) && (i < screen_info->client_count);
             c2 = c2->prev, i++)
         {
-            if (clientSelectMask (c2, mask, WINDOW_REGULAR_FOCUSABLE))
+            if (clientSelectMask (c2, c, mask, type))
             {
                 return c2;
             }
@@ -500,11 +526,11 @@ clientUpdateFocus (ScreenInfo *screen_info, Client * c, unsigned short flags)
         {
             TRACE ("Un-setting WM_STATE_DEMANDS_ATTENTION flag on \"%s\" (0x%lx)", c->name, c->window);
             FLAG_UNSET (c->flags, CLIENT_FLAG_DEMANDS_ATTENTION);
+            clientSetNetState (c);
         }
         clientAdjustFullscreenLayer (c, TRUE);
         frameQueueDraw (c, FALSE);
         clientUpdateOpacity (c);
-        clientSetNetState (c);
     }
     if (c2)
     {
@@ -608,8 +634,8 @@ clientInitFocusFlag (Client * c)
 {
     ScreenInfo *screen_info;
     Client *c2;
-    GList *index;
-    int workspace;
+    GList *list;
+    guint workspace;
 
     g_return_if_fail (c != NULL);
     TRACE ("entering clientSetFocus");
@@ -621,9 +647,9 @@ clientInitFocusFlag (Client * c)
 
     screen_info = c->screen_info;
     workspace = c->win_workspace;
-    for (index = screen_info->windows_stack; index; index = g_list_next (index))
+    for (list = screen_info->windows_stack; list; list = g_list_next (list))
     {
-        c2 = (Client *) index->data;
+        c2 = (Client *) list->data;
         if ((c2->win_workspace == workspace) && FLAG_TEST (c2->xfwm_flags, XFWM_FLAG_FOCUS))
         {
             FLAG_UNSET (c2->xfwm_flags, XFWM_FLAG_FOCUS);
@@ -716,7 +742,7 @@ void
 clientGrabMouseButtonForAll (ScreenInfo *screen_info)
 {
     Client *c;
-    int i;
+    guint i;
 
     g_return_if_fail (screen_info != NULL);
     TRACE ("entering clientGrabMouseButtonForAll");
@@ -732,7 +758,7 @@ void
 clientUngrabMouseButtonForAll (ScreenInfo *screen_info)
 {
     Client *c;
-    int i;
+    guint i;
 
     g_return_if_fail (screen_info != NULL);
     TRACE ("entering clientUngrabMouseButtonForAll");
