@@ -1,4 +1,4 @@
-/*      $Id: netwm.c 29115 2009-01-07 22:18:25Z olivier $
+/*      $Id$
 
         This program is free software; you can redistribute it and/or modify
         it under the terms of the GNU General Public License as published by
@@ -16,7 +16,7 @@
         MA 02110-1301, USA.
 
 
-        xfwm4    - (c) 2002-2008 Olivier Fourdan
+        xfwm4    - (c) 2002-2009 Olivier Fourdan
 
  */
 
@@ -728,7 +728,6 @@ clientUpdateFullscreenState (Client * c)
 {
     ScreenInfo *screen_info;
     DisplayInfo *display_info;
-    XWindowChanges wc;
     int layer;
 
     g_return_if_fail (c != NULL);
@@ -740,48 +739,19 @@ clientUpdateFullscreenState (Client * c)
 
     if (FLAG_TEST (c->flags, CLIENT_FLAG_FULLSCREEN))
     {
-        GdkRectangle rect;
-        gint monitor_nbr;
-        int cx, cy;
-
-        cx = frameX (c) + (frameWidth (c) / 2);
-        cy = frameY (c) + (frameHeight (c) / 2);
-
-        monitor_nbr = find_monitor_at_point (c->screen_info->gscr, cx, cy);
-        gdk_screen_get_monitor_geometry (c->screen_info->gscr, monitor_nbr, &rect);
-
         c->fullscreen_old_x = c->x;
         c->fullscreen_old_y = c->y;
         c->fullscreen_old_width = c->width;
         c->fullscreen_old_height = c->height;
         c->fullscreen_old_layer = c->win_layer;
-
-        wc.x = rect.x;
-        wc.y = rect.y;
-        wc.width = rect.width;
-        wc.height = rect.height;
         layer = WIN_LAYER_FULLSCREEN;
     }
     else
     {
-        wc.x = c->fullscreen_old_x;
-        wc.y = c->fullscreen_old_y;
-        wc.width = c->fullscreen_old_width;
-        wc.height = c->fullscreen_old_height;
         layer = c->fullscreen_old_layer;
     }
-    if (FLAG_TEST (c->xfwm_flags, XFWM_FLAG_MANAGED))
-    {
-        clientConfigure (c, &wc, CWX | CWY | CWWidth | CWHeight, CFG_FORCE_REDRAW);
-    }
-    else
-    {
-        c->x = wc.x;
-        c->y = wc.y;
-        c->height = wc.height;
-        c->width = wc.width;
-    }
     clientSetLayer (c, layer);
+    clientUpdateFullscreenSize (c);
 
     /* Fullscreen has no decoration at all, update NET_FRAME_EXTENTS accordingly */
     setNetFrameExtents (display_info,
@@ -822,9 +792,6 @@ clientGetNetWmType (Client * c)
             case WIN_LAYER_DOCK:
                 c->type_atom = display_info->atoms[NET_WM_WINDOW_TYPE_DOCK];
                 break;
-            case WIN_LAYER_NORMAL:
-                c->type_atom = display_info->atoms[NET_WM_WINDOW_TYPE_NORMAL];
-                break;
             default:
                 if (c->transient_for != None)
                 {
@@ -1281,8 +1248,10 @@ clientWindowType (Client * c)
             c->type = WINDOW_DIALOG;
             c->initial_layer = WIN_LAYER_NORMAL;
             /* Treat DIALOG without transient_for set as transient for group */
-            if (c->transient_for == None)
+            if ((c->transient_for == None) || (!clientGetTransient (c)))
             {
+                TRACE ("Invalid transient 0x%lx specified for dialog window 0x%lx (%s)",
+                       c->transient_for, c->window, c->name);
                 c->transient_for = c->screen_info->xroot;
             }
         }
@@ -1300,8 +1269,10 @@ clientWindowType (Client * c)
             c->type = WINDOW_UTILITY;
             c->initial_layer = WIN_LAYER_NORMAL;
             /* Treat UTILITY without transient_for set as transient for group */
-            if (c->transient_for == None)
+            if ((c->transient_for == None) || (!clientGetTransient (c)))
             {
+                TRACE ("Invalid transient 0x%lx specified for utility window 0x%lx (%s)",
+                       c->transient_for, c->window, c->name);
                 c->transient_for = c->screen_info->xroot;
             }
         }
@@ -1335,15 +1306,18 @@ clientWindowType (Client * c)
         if ((c2) && (c->initial_layer < c2->win_layer))
         {
             c->initial_layer = c2->win_layer;
-            TRACE ("Applied layer is %i", c->initial_layer);
+            TRACE ("Applied layer is %lu", c->initial_layer);
         }
         FLAG_UNSET (c->xfwm_flags, XFWM_FLAG_HAS_HIDE);
     }
-    if ((old_type != c->type) || (c->initial_layer != c->win_layer))
+    if (!FLAG_TEST (c->flags, CLIENT_FLAG_ABOVE|CLIENT_FLAG_BELOW|CLIENT_FLAG_FULLSCREEN))
     {
-        TRACE ("setting layer %i", c->initial_layer);
-        clientSetLayer (c, c->initial_layer);
-        clientSetNetState (c);
+        if ((old_type != c->type) || (c->initial_layer != c->win_layer))
+        {
+            TRACE ("setting layer %lu", c->initial_layer);
+            clientSetLayer (c, c->initial_layer);
+            clientSetNetState (c);
+        }
     }
 }
 
@@ -1392,6 +1366,49 @@ clientSetNetActiveWindow (ScreenInfo *screen_info, Client *c, guint32 timestamp)
                      PropModeReplace, (unsigned char *) data, 2);
 }
 
+void
+clientHandleNetActiveWindow (Client *c, guint32 timestamp, gboolean source_is_application)
+{
+    DisplayInfo *display_info;
+    ScreenInfo *screen_info;
+    guint32 ev_time, current_time;
+    Client *focused;
+
+    g_return_if_fail (c != NULL);
+    TRACE ("entering clientHandleNetActiveWindow");
+
+    screen_info = c->screen_info;
+    display_info = screen_info->display_info;
+    ev_time = myDisplayGetTime (display_info, timestamp);
+
+    if (source_is_application)
+    {
+        current_time = myDisplayGetLastUserTime (display_info);
+
+        TRACE ("Time of event received is %u, current XServer time is %u", (guint32) ev_time, (guint32) current_time);
+        if ((screen_info->params->prevent_focus_stealing) && TIMESTAMP_IS_BEFORE((guint32) ev_time, (guint32) current_time))
+        {
+            focused = clientGetFocus ();
+            /* We do not want to set the demand attention flag if the window is focused though */
+            if (c != focused)
+            {
+                TRACE ("Setting WM_STATE_DEMANDS_ATTENTION flag on \"%s\" (0x%lx)", c->name, c->window);
+                FLAG_SET (c->flags, CLIENT_FLAG_DEMANDS_ATTENTION);
+                clientSetNetState (c);
+            }
+        }
+        else
+        {
+            clientActivate (c, ev_time, source_is_application);
+        }
+    }
+    else
+    {
+        /* The request is either from a pager or an older client, use the most accurate timestamp */
+        clientActivate (c, getXServerTime (display_info), source_is_application);
+    }
+}
+
 static gboolean
 ping_timeout_cb (gpointer data)
 {
@@ -1428,7 +1445,7 @@ void
 clientReceiveNetWMPong (ScreenInfo *screen_info, guint32 timestamp)
 {
     Client *c;
-    int i;
+    guint i;
 
     g_return_if_fail (screen_info != NULL);
     g_return_if_fail (timestamp != CurrentTime);
@@ -1476,3 +1493,74 @@ clientSendNetWMPing (Client *c, guint32 timestamp)
                             (gpointer) c, NULL);
     return (TRUE);
 }
+
+gboolean
+clientGetUserTime (Client * c)
+{
+    ScreenInfo *screen_info;
+    DisplayInfo *display_info;
+
+    g_return_val_if_fail (c != NULL, FALSE);
+    g_return_val_if_fail (c->window != None, FALSE);
+
+    screen_info = c->screen_info;
+    display_info = screen_info->display_info;
+
+    /*
+     * We can use "c->user_time_win" safely here because this will be
+     * the same as "c->window" if the app does not support the protocol
+     * NET_WM_USER_TIME_WINDOW
+     */
+
+    if (getNetWMUserTime (display_info, c->user_time_win, &c->user_time))
+    {
+        guint32 last_user_time;
+
+        last_user_time = myDisplayGetLastUserTime (display_info);
+        if (c->user_time && TIMESTAMP_IS_BEFORE(last_user_time, c->user_time))
+        {
+            myDisplaySetLastUserTime (display_info, c->user_time);
+        }
+        FLAG_SET (c->flags, CLIENT_FLAG_HAS_USER_TIME);
+
+        return TRUE;
+    }
+
+    return FALSE;
+}
+
+void
+clientAddUserTimeWin (Client * c)
+{
+    ScreenInfo *screen_info;
+    DisplayInfo *display_info;
+
+    g_return_if_fail (c != NULL);
+    g_return_if_fail (c->window != None);
+
+    screen_info = c->screen_info;
+    display_info = screen_info->display_info;
+
+    if ((c->user_time_win != None) && (c->user_time_win != c->window))
+    {
+        XSelectInput (display_info->dpy, c->user_time_win, PropertyChangeMask);
+    }
+}
+
+void
+clientRemoveUserTimeWin (Client * c)
+{
+    ScreenInfo *screen_info;
+    DisplayInfo *display_info;
+
+    g_return_if_fail (c != NULL);
+    g_return_if_fail (c->window != None);
+
+    screen_info = c->screen_info;
+    display_info = screen_info->display_info;
+
+    if ((c->user_time_win != None) && (c->user_time_win != c->window))
+    {
+        XSelectInput (display_info->dpy, c->user_time_win, NoEventMask);
+    }
+}
