@@ -1,4 +1,4 @@
-/*      $Id: parserc.c 29303 2009-01-23 22:36:35Z olivier $
+/*      $Id$
 
         This program is free software; you can redistribute it and/or modify
         it under the terms of the GNU General Public License as published by
@@ -155,7 +155,7 @@ setValue (const gchar * lvalue, const gchar *rvalue, Settings *rc)
     return FALSE;
 }
 
-gboolean
+static gboolean
 setGValue (const gchar * lvalue, const GValue *rvalue, Settings *rc)
 {
     gint i;
@@ -213,7 +213,7 @@ setStringValue (const gchar * lvalue, const gchar *value, Settings *rc)
 {
     GValue tmp_val = {0, };
     g_value_init(&tmp_val, G_TYPE_STRING);
-    g_value_set_string(&tmp_val, value);
+    g_value_set_static_string(&tmp_val, value);
     return setGValue (lvalue, &tmp_val, rc);
 }
 
@@ -300,7 +300,11 @@ getStringValue (const gchar *option, Settings *rc)
         {
             if (rc[i].value == NULL)
                 return NULL;
-            g_return_val_if_fail(G_VALUE_TYPE(rc[i].value) == G_TYPE_STRING, NULL);
+            if (G_VALUE_TYPE(rc[i].value) != G_TYPE_STRING)
+            {
+                g_warning ("Option \"%s\" in not of expected type string", option);
+                return NULL;
+            }
             return g_value_get_string(rc[i].value);
         }
     }
@@ -322,7 +326,11 @@ getIntValue (const gchar *option, Settings *rc)
         {
             if (rc[i].value == NULL)
                 return 0;
-            g_return_val_if_fail(G_VALUE_TYPE(rc[i].value) == G_TYPE_INT, 0);
+            if (G_VALUE_TYPE(rc[i].value) != G_TYPE_INT)
+            {
+                g_warning ("Option \"%s\" in not of expected type int", option);
+                return 0;
+            }
             return g_value_get_int(rc[i].value);
         }
     }
@@ -344,7 +352,11 @@ getBoolValue (const gchar *option, Settings *rc)
         {
             if (rc[i].value == NULL)
                 return FALSE;
-            g_return_val_if_fail(G_VALUE_TYPE(rc[i].value) == G_TYPE_BOOLEAN, FALSE);
+            if (G_VALUE_TYPE(rc[i].value) != G_TYPE_BOOLEAN)
+            {
+                g_warning ("Option \"%s\" in not of expected type boolean", option);
+                return FALSE;
+            }
             return g_value_get_boolean(rc[i].value);
         }
     }
