@@ -302,29 +302,10 @@ getTitleShadow (Settings *rc, const gchar * name)
     return TITLE_SHADOW_NONE;
 }
 
-static int
-getFontHeight (const PangoFontDescription *desc, PangoContext *context)
-{
-    PangoFontMetrics *metrics;
-    PangoLanguage *language;
-    int height;
-
-    g_return_val_if_fail (desc, 0);
-    g_return_val_if_fail (context, 0);
-
-    language = pango_context_get_language (context);
-    metrics = pango_context_get_metrics (context, desc, language);
-    height = PANGO_PIXELS (pango_font_metrics_get_ascent (metrics) +
-                           pango_font_metrics_get_descent (metrics));
-    pango_font_metrics_unref (metrics);
-
-    return height;
-}
-
-
 static void
 loadTheme (ScreenInfo *screen_info, Settings *rc)
 {
+
     static const char *side_names[] = {
         "left",
         "right",
@@ -358,6 +339,58 @@ loadTheme (ScreenInfo *screen_info, Settings *rc)
         "toggled-pressed"
     };
 
+    static const char *ui_part[] = {
+        "fg",
+        "fg",
+        "dark",
+        "dark",
+        "fg",
+        "fg",
+        "bg",
+        "light",
+        "dark",
+        "mid",
+        "bg",
+        "light",
+        "dark",
+        "mid",
+        "bg",
+        "light",
+        "dark",
+        "mid",
+        "bg",
+        "light",
+        "dark",
+        "mid",
+        NULL
+    };
+
+    static const char *ui_state[] = {
+        "selected",
+        "insensitive",
+        "selected",
+        "insensitive",
+        "normal",
+        "normal",
+        "selected",
+        "selected",
+        "selected",
+        "selected",
+        "normal",
+        "normal",
+        "normal",
+        "normal",
+        "insensitive",
+        "insensitive",
+        "insensitive",
+        "insensitive",
+        "normal",
+        "normal",
+        "normal",
+        "normal",
+        NULL
+    };
+
     gchar imagename[30];
     GValue tmp_val = { 0, };
     DisplayInfo *display_info;
@@ -374,30 +407,17 @@ loadTheme (ScreenInfo *screen_info, Settings *rc)
 
     desc = NULL;
     context = NULL;
+    i = 0;
 
-    setStringValue (rc[0].option, getUIStyle (widget, "fg",    "selected"), rc);
-    setStringValue (rc[1].option, getUIStyle (widget, "fg",    "insensitive"), rc);
-    setStringValue (rc[2].option, getUIStyle (widget, "dark",  "selected"), rc);
-    setStringValue (rc[3].option, getUIStyle (widget, "dark",  "insensitive"), rc);
-    setStringValue (rc[4].option, getUIStyle (widget, "fg",    "normal"), rc);
-    setStringValue (rc[5].option, getUIStyle (widget, "fg",    "normal"), rc);
-    setStringValue (rc[6].option, getUIStyle (widget, "bg",    "selected"), rc);
-    setStringValue (rc[7].option, getUIStyle (widget, "light", "selected"), rc);
-    setStringValue (rc[8].option, getUIStyle (widget, "dark",  "selected"), rc);
-    setStringValue (rc[9].option, getUIStyle (widget, "mid",   "selected"), rc);
-    setStringValue (rc[10].option, getUIStyle (widget, "bg",    "normal"), rc);
-    setStringValue (rc[11].option, getUIStyle (widget, "light", "normal"), rc);
-    setStringValue (rc[12].option, getUIStyle (widget, "dark",  "normal"), rc);
-    setStringValue (rc[13].option, getUIStyle (widget, "mid",   "normal"), rc);
-    setStringValue (rc[14].option, getUIStyle (widget, "bg",    "insensitive"), rc);
-    setStringValue (rc[15].option, getUIStyle (widget, "light", "insensitive"), rc);
-    setStringValue (rc[16].option, getUIStyle (widget, "dark",  "insensitive"), rc);
-    setStringValue (rc[17].option, getUIStyle (widget, "mid",   "insensitive"), rc);
-    setStringValue (rc[18].option, getUIStyle (widget, "bg",    "normal"), rc);
-    setStringValue (rc[19].option, getUIStyle (widget, "light", "normal"), rc);
-    setStringValue (rc[20].option, getUIStyle (widget, "dark",  "normal"), rc);
-    setStringValue (rc[21].option, getUIStyle (widget, "mid",   "normal"), rc);
+    while (ui_part[i] && ui_state[i])
+    {
+        gchar *color;
 
+        color = getUIStyle  (widget, ui_part[i], ui_state[i]);
+        setStringValue (rc[i].option, color, rc);
+        g_free (color);
+        ++i;
+    }
 
     theme = getThemeDir (getStringValue ("theme", rc), THEMERC);
     parseRc (THEMERC, theme, rc);
@@ -435,8 +455,8 @@ loadTheme (ScreenInfo *screen_info, Settings *rc)
         if (desc)
         {
             gtk_widget_modify_font (widget, desc);
-            screen_info->font_height = getFontHeight (desc, context);
             pango_font_description_free (desc);
+            myScreenUpdateFontHeight (screen_info);
         }
     }
 
@@ -582,6 +602,8 @@ loadKeyBindings (ScreenInfo *screen_info)
     parseShortcut (screen_info, KEY_SHADE_WINDOW, "shade_window_key", shortcuts);
     parseShortcut (screen_info, KEY_SHOW_DESKTOP, "show_desktop_key", shortcuts);
     parseShortcut (screen_info, KEY_STICK_WINDOW, "stick_window_key", shortcuts);
+    parseShortcut (screen_info, KEY_SWITCH_APPLICATION, "switch_application_key", shortcuts);
+    parseShortcut (screen_info, KEY_SWITCH_WINDOW, "switch_window_key", shortcuts);
     parseShortcut (screen_info, KEY_TOGGLE_ABOVE, "above_key", shortcuts);
     parseShortcut (screen_info, KEY_TOGGLE_FULLSCREEN, "fullscreen_key", shortcuts);
     parseShortcut (screen_info, KEY_UP_WORKSPACE, "up_workspace_key", shortcuts);
@@ -641,6 +663,7 @@ loadSettings (ScreenInfo *screen_info)
         {"button_spacing", NULL, G_TYPE_INT, TRUE},
         {"click_to_focus", NULL, G_TYPE_BOOLEAN, TRUE},
         {"focus_delay", NULL, G_TYPE_INT, TRUE},
+        {"cycle_apps_only", NULL, G_TYPE_BOOLEAN, TRUE},
         {"cycle_draw_frame", NULL, G_TYPE_BOOLEAN, TRUE},
         {"cycle_hidden", NULL, G_TYPE_BOOLEAN, TRUE},
         {"cycle_minimum", NULL, G_TYPE_BOOLEAN, TRUE},
@@ -707,7 +730,6 @@ loadSettings (ScreenInfo *screen_info)
 
     TRACE ("entering loadSettings");
 
-    loadRcData (screen_info, rc);
     loadRcData (screen_info, rc);
     loadXfconfData (screen_info, rc);
     loadTheme (screen_info, rc);
@@ -723,6 +745,8 @@ loadSettings (ScreenInfo *screen_info)
         getBoolValue ("box_move", rc);
     screen_info->params->click_to_focus =
         getBoolValue ("click_to_focus", rc);
+    screen_info->params->cycle_apps_only =
+        getBoolValue ("cycle_apps_only", rc);
     screen_info->params->cycle_minimum =
         getBoolValue ("cycle_minimum", rc);
     screen_info->params->cycle_draw_frame =
@@ -838,10 +862,10 @@ loadSettings (ScreenInfo *screen_info)
         screen_info->params->double_click_action = DBL_CLICK_ACTION_NONE;
     }
 
-    if (screen_info->workspace_count < 0)
+    if (screen_info->workspace_count == 0)
     {
-        gint workspace_count;
-        workspace_count = MAX (getIntValue ("workspace_count", rc), 0);
+        guint workspace_count;
+        workspace_count = MAX ((guint) getIntValue ("workspace_count", rc), 1);
         workspaceSetCount (screen_info, workspace_count);
     }
 
@@ -917,7 +941,7 @@ reloadScreenSettings (ScreenInfo *screen_info, int mask)
 {
     g_return_val_if_fail (screen_info, FALSE);
 
-    unloadTheme (screen_info);
+    unloadSettings (screen_info);
     if (!loadSettings (screen_info))
     {
         return FALSE;
@@ -963,7 +987,7 @@ initSettings (ScreenInfo *screen_info)
     DisplayInfo *display_info;
     char **names;
     long val;
-    int i;
+    guint i;
 
     g_return_val_if_fail (screen_info, FALSE);
 
@@ -1079,25 +1103,25 @@ cb_xfwm4_channel_property_changed(XfconfChannel *channel, const gchar *property_
                 }
                 else
                 {
-                    g_warning("The property '%s' is not supported", property_name);
+                    g_warning("The property '%s' of type string is not supported", property_name);
                 }
                 break;
             case G_TYPE_INT:
                 if (!strcmp (name, "raise_delay"))
                 {
-                    screen_info->params->raise_delay = g_value_get_int (value);
+                    screen_info->params->raise_delay = CLAMP (g_value_get_int (value), 5, 2000);
                 }
                 else if (!strcmp (name, "focus_delay"))
                 {
-                    screen_info->params->focus_delay = g_value_get_int (value);
+                    screen_info->params->focus_delay = CLAMP (g_value_get_int (value), 5, 2000);
                 }
                 else if (!strcmp (name, "snap_width"))
                 {
-                    screen_info->params->snap_width = g_value_get_int (value);
+                    screen_info->params->snap_width = CLAMP (g_value_get_int (value), 5, 100);
                 }
                 else if (!strcmp (name, "wrap_resistance"))
                 {
-                    screen_info->params->wrap_resistance = g_value_get_int (value);
+                    screen_info->params->wrap_resistance = CLAMP (g_value_get_int (value), 5, 100);
                 }
                 else if (!strcmp (name, "margin_left"))
                 {
@@ -1121,31 +1145,31 @@ cb_xfwm4_channel_property_changed(XfconfChannel *channel, const gchar *property_
                 }
                 else if (!strcmp (name, "frame_opacity"))
                 {
-                    screen_info->params->frame_opacity = g_value_get_int(value);
+                    screen_info->params->frame_opacity = CLAMP (g_value_get_int(value), 0, 100);
                     reloadScreenSettings (screen_info, UPDATE_FRAME);
                 }
                 else if (!strcmp (name, "inactive_opacity"))
                 {
-                    screen_info->params->inactive_opacity = g_value_get_int(value);
+                    screen_info->params->inactive_opacity = CLAMP (g_value_get_int(value), 0, 100);
                     reloadScreenSettings (screen_info, UPDATE_FRAME);
                     clientUpdateAllOpacity (screen_info);
                 }
                 else if (!strcmp (name, "move_opacity"))
                 {
-                    screen_info->params->move_opacity = g_value_get_int (value);
+                    screen_info->params->move_opacity = CLAMP (g_value_get_int(value), 0, 100);
                 }
                 else if (!strcmp (name, "resize_opacity"))
                 {
-                    screen_info->params->resize_opacity = g_value_get_int (value);
+                    screen_info->params->resize_opacity = CLAMP (g_value_get_int(value), 0, 100);
                 }
                 else if (!strcmp (name, "popup_opacity"))
                 {
-                    screen_info->params->popup_opacity = g_value_get_int (value);
+                    screen_info->params->popup_opacity = CLAMP (g_value_get_int(value), 0, 100);
                     reloadScreenSettings (screen_info, UPDATE_FRAME);
                 }
                 else if (!strcmp (name, "placement_ratio"))
                 {
-                    screen_info->params->placement_ratio = g_value_get_int (value);
+                    screen_info->params->placement_ratio = CLAMP (g_value_get_int(value), 0, 100);
                 }
                 else if ((!strcmp (name, "button_offset"))
                       || (!strcmp (name, "button_spacing"))
@@ -1164,7 +1188,7 @@ cb_xfwm4_channel_property_changed(XfconfChannel *channel, const gchar *property_
                 }
                 else
                 {
-                    g_warning("The property '%s' is not supported", property_name);
+                    g_warning("The property '%s' of type int is not supported", property_name);
                 }
                 break;
             case G_TYPE_BOOLEAN:
@@ -1240,6 +1264,10 @@ cb_xfwm4_channel_property_changed(XfconfChannel *channel, const gchar *property_
                 {
                     screen_info->params->cycle_workspaces = g_value_get_boolean (value);
                 }
+                else if (!strcmp (name, "cycle_apps_only"))
+                {
+                    screen_info->params->cycle_apps_only = g_value_get_boolean (value);
+                }
                 else if (!strcmp (name, "focus_hint"))
                 {
                     screen_info->params->focus_hint = g_value_get_boolean (value);
@@ -1309,7 +1337,7 @@ cb_xfwm4_channel_property_changed(XfconfChannel *channel, const gchar *property_
                 }
                 else
                 {
-                    g_warning("The property '%s' is not supported", property_name);
+                    g_warning("The property '%s' of type boolean is not supported", property_name);
                 }
                 break;
             default:
@@ -1329,8 +1357,15 @@ cb_xfwm4_channel_property_changed(XfconfChannel *channel, const gchar *property_
 static void
 cb_keys_changed (GdkKeymap *keymap, ScreenInfo *screen_info)
 {
+    /* Recompute modifiers mask in case it changed */
     initModifiers (myScreenGetXDisplay (screen_info));
-    reloadSettings (screen_info->display_info, UPDATE_BUTTON_GRABS);
+
+    /* Ungrab/regrab shortcuts */
+    myScreenUngrabKeys (screen_info);
+    myScreenGrabKeys (screen_info);
+
+    /* Uupdate all grabs for mouse buttons */
+    clientUpdateAllFrames (screen_info, UPDATE_BUTTON_GRABS);
 }
 
 static void
@@ -1401,20 +1436,20 @@ cb_shortcut_removed (XfceShortcutsProvider *provider, const gchar *shortcut,
 }
 
 static void
-parseShortcut (ScreenInfo *screen_info, int index, const gchar *name,
+parseShortcut (ScreenInfo *screen_info, int id, const gchar *name,
                GList *shortcuts)
 {
     Display *dpy;
     const gchar *shortcut;
 
     g_return_if_fail (screen_info);
-    g_return_if_fail (index >= 0 && index < KEY_COUNT);
+    g_return_if_fail (id >= 0 && id < KEY_COUNT);
 
     dpy = myScreenGetXDisplay (screen_info);
     shortcut = getShortcut (name, shortcuts);
-    parseKeyString (dpy, &screen_info->params->keys[index], shortcut);
+    parseKeyString (dpy, &screen_info->params->keys[id], shortcut);
 
-    screen_info->params->keys[index].internal_name = g_strdup (name);
+    screen_info->params->keys[id].internal_name = g_strdup (name);
 }
 
 static const gchar *
